@@ -9,6 +9,7 @@ module Helpers
     def spawn_process(*args)
       env = {
         "LC_ALL" => "C",
+        "PGCLIENTENCODING" => "UTF-8",
       }
       output_read, output_write = IO.pipe
       error_read, error_write = IO.pipe
@@ -77,6 +78,7 @@ module Helpers
     attr_reader :dir
     attr_reader :host
     attr_reader :port
+    attr_reader :user
     attr_reader :replication_user
     attr_reader :replication_password
     def initialize(base_dir)
@@ -96,11 +98,13 @@ module Helpers
       @running
     end
 
-    def initdb(shared_preload_libraries: [])
-      @dir = File.join(@base_dir, "db")
+    def initdb(shared_preload_libraries: [],
+               db_path: "db",
+               port: 15432)
+      @dir = File.join(@base_dir, db_path)
       @log_path = File.join(@dir, "log", @log_base_name)
       socket_dir = File.join(@dir, "socket")
-      @port = 15432
+      @port = port
       @replication_user = "replicator"
       run_command("initdb",
                   "--locale", "C",
@@ -139,7 +143,7 @@ module Helpers
                   @replication_user)
     end
 
-    def init_replication(primary)
+    def init_replication(primary, shared_preload_libraries: [])
       @dir = File.join(@base_dir, "db-standby")
       @log_path = File.join(@dir, "log", @log_base_name)
       @port = primary.port + 1
@@ -158,7 +162,8 @@ module Helpers
       end
       conf = File.read(postgresql_conf)
       conf = conf.gsub(/^shared_preload_libraries = '(.*?)'/) do
-        "shared_preload_libraries = '#{$1},pgroonga_wal_applier'"
+        libraries = [$1, *shared_preload_libraries].join(",")
+        "shared_preload_libraries = '#{libraries}'"
       end
       File.write(postgresql_conf, conf)
     end
@@ -169,7 +174,7 @@ module Helpers
                     "-w",
                     "-D", @dir)
       rescue => error
-        error.message << "\nPostgreSQL log:\n#{File.read(@log_path)}"
+        error.message << "\nPostgreSQL log:\n#{read_log}"
         raise
       end
       loop do
@@ -192,14 +197,16 @@ module Helpers
     end
 
     def psql(db, sql)
-      run_command("psql",
-                  "--host", @host,
-                  "--port", @port.to_s,
-                  "--username", @user,
-                  "--dbname", db,
-                  "--echo-all",
-                  "--no-psqlrc",
-                  "--command", sql)
+      output, error = run_command("psql",
+                                  "--host", @host,
+                                  "--port", @port.to_s,
+                                  "--username", @user,
+                                  "--dbname", db,
+                                  "--echo-all",
+                                  "--no-psqlrc",
+                                  "--command", sql)
+      output = normalize_output(output)
+      [output, error]
     end
 
     def groonga(*command_line)
@@ -210,10 +217,27 @@ module Helpers
       JSON.parse(output)
     end
 
+    def read_log
+      return "" unless File.exist?(@log_path)
+      File.read(@log_path)
+    end
+
     private
     def windows?
       /mingw|mswin|cygwin/.match?(RUBY_PLATFORM)
     end
+
+    def normalize_output(output)
+      normalized_output = ""
+      output.each_line do |line|
+        case line.chomp
+        when "SET", "CREATE EXTENSION"
+          next
+        end
+        normalized_output << line
+      end
+      normalized_output
+    end
   end
 
   module Sandbox
@@ -269,7 +293,13 @@ module Helpers
     end
 
     def teardown_tmp_dir
-      FileUtils.rm_rf(@tmp_dir)
+      debug_dir = ENV["PGROONGA_TEST_DEBUG_DIR"]
+      if debug_dir and File.exist?(@tmp_dir)
+        FileUtils.rm_rf(debug_dir)
+        FileUtils.mv(@tmp_dir, debug_dir)
+      else
+        FileUtils.rm_rf(@tmp_dir)
+      end
     end
 
     def setup_db
@@ -293,9 +323,16 @@ module Helpers
     def teardown_db
     end
 
+    def shared_preload_libraries_standby
+      []
+    end
+
     def setup_standby_db
       @postgresql_standby = PostgreSQL.new(@tmp_dir)
-      @postgresql_standby.init_replication(@postgresql) do |conf|
+      options = {
+        shared_preload_libraries: shared_preload_libraries_standby,
+      }
+      @postgresql_standby.init_replication(@postgresql, **options) do |conf|
         conf.puts(additional_standby_configurations)
       end
       @postgresql_standby.start
@@ -306,7 +343,32 @@ module Helpers
     end
 
     def teardown_standby_db
-      @postgresql_standby.stop
+      @postgresql_standby.stop if @postgresql_standby
+    end
+
+    def setup_reference_db
+      @postgresql_reference = PostgreSQL.new(@tmp_dir)
+      options = {
+        db_path: "db-reference",
+        port: 25432,
+        shared_preload_libraries: reference_shared_preload_libraries,
+      }
+      @postgresql_reference.initdb(**options) do |conf|
+        conf.puts(additional_reference_configurations)
+      end
+      @postgresql_reference.start
+    end
+
+    def reference_shared_preload_libraries
+      []
+    end
+
+    def additional_reference_configurations
+      ""
+    end
+
+    def teardown_reference_db
+      @postgresql_reference.stop if @postgresql_reference
     end
 
     def start_postgres
@@ -323,14 +385,18 @@ module Helpers
     end
 
     def teardown_postgres
-      stop_postgres
+      stop_postgres if @postgresql
+    end
+
+    def create_db(postgresql, db_name)
+      postgresql.psql("postgres", "CREATE DATABASE #{db_name}")
+      postgresql.psql(db_name, "CREATE EXTENSION pgroonga")
+      postgresql.psql(db_name, "CHECKPOINT")
     end
 
     def setup_test_db
       @test_db_name = "test"
-      psql("postgres", "CREATE DATABASE #{@test_db_name}")
-      run_sql("CREATE EXTENSION pgroonga")
-      run_sql("CHECKPOINT")
+      create_db(@postgresql, @test_db_name)
       result, = run_sql("SELECT oid FROM pg_catalog.pg_database " +
                         "WHERE datname = current_database()")
       oid = result.lines[3].strip
@@ -339,5 +405,12 @@ module Helpers
 
     def teardown_test_db
     end
+
+    def setup_reference_test_db
+      create_db(@postgresql_reference, @test_db_name)
+    end
+
+    def teardown_reference_test_db
+    end
   end
 end
